# -*- coding: utf-8 -*-
import coloredlogs
import click
import click_completion
import crayons
import ruamel.yaml

from .version import __version__
from . import commands
from cg.invoice import invoice as invoice_cli
from cg.queue import cli as queue_cli

# enable shell completion
click_completion.init()


@click.group(invoke_without_command=True)
@click.version_option(prog_name=crayons.yellow('cg'), version=__version__)
@click.option('-h', '--help', is_flag=True, help="Show this message then exit.")
@click.option('-l', '--log-level', default='INFO')
@click.option('-c', '--config', type=click.File('r'), required=True)
@click.pass_context
def cli(context, help, log_level, config):
    """Central place for interacting with CG apps."""
    coloredlogs.install(level=log_level)
    if help:
        click.echo(context.get_help())
    context.obj = ruamel.yaml.safe_load(config)


cli.add_command(commands.mip_config)
cli.add_command(commands.reruns)
cli.add_command(commands.update)
cli.add_command(commands.check)
cli.add_command(commands.mip_config)
cli.add_command(commands.mip_panel)
cli.add_command(commands.start)
cli.add_command(queue_cli.queue)
cli.add_command(commands.coverage)
cli.add_command(commands.genotypes)
cli.add_command(commands.qc)
cli.add_command(commands.visualize)
cli.add_command(commands.delivery_report)
cli.add_command(invoice_cli)
cli.add_command(commands.add)
cli.add_command(commands.validate)
cli.add_command(commands.observations)
cli.add_command(commands.auto_start)
cli.add_command(commands.keep)



# -*- coding: utf-8 -*-
import logging

import click
from dateutil.parser import parse as parse_date

from .api import QueueApi

log = logging.getLogger(__name__)


@click.group()
@click.option('-d', '--database', type=click.Path(), help='path to database')
@click.pass_context
def queue(context, database):
    """Interface to analysis queue."""
    database = database or context.obj['queue']['database']
    api = QueueApi(database)
    api.connect()
    context.obj = dict(api=api)


@queue.command()
@click.pass_context
def init(context):
    """Setup the database."""
    context.obj['api'].db.create_all()


@queue.command()
@click.option('-a', '--last-analyzed', help='last analysis date')
@click.option('-p', '--prioritize', is_flag=True, help='prioritize case')
@click.argument('case_id')
@click.pass_context
def add(context, case_id, last_analyzed=None, prioritize=None):
    """Add a new case to be analyzed."""
    old_case = context.obj['api'].case(case_id)
    if old_case:
        click.echo("case already exists: {}".format(old_case['case_id']))
    else:
        last_analyzed_at = parse_date(last_analyzed) if last_analyzed else None
        new_case = dict(case_id=case_id, last_analyzed_at=last_analyzed_at,
                        is_prioritized=prioritize)
        case_obj = context.obj['api'].save_case(new_case)
        click.echo("new case created: {case.case_id} ({case.id})".format(case=case_obj))


@queue.command()
@click.option('-l', '--limit', default=20, help='how many cases to display')
@click.pass_context
def cases(context, limit):
    """List cases in order of higest priority."""
    all_cases = context.obj['api'].cases()
    ready_cases = (case_obj for case_obj in all_cases if case_obj.is_ready())
    for index, case_obj in enumerate(ready_cases):
        if index > limit:
            break

        click.echo(case_obj.case_id)



# -*- coding: utf-8 -*-
from __future__ import division
import datetime
import hashlib
import logging

from cgadmin.store import api, models
from cgadmin.report.core import export_report as export_report_api

log = logging.getLogger(__name__)


class Application(api.AdminDatabase):

    """Admin database API.

    Args:
        config (dict): CLI config object
    """

    def __init__(self, config):
        super(Application, self).__init__(config['cgadmin']['database'])

    def map_apptags(self, apptags):
        """Map application tags with latest versions.

        Args:
            admin_db (sqlservice.SQLClient)
        """
        apptag_map = {}
        for apptag_id in apptags:
            latest_version = self.latest_version(apptag_id)
            apptag_map[apptag_id] = latest_version.version
        return apptag_map

    def customer(self, customer_id):
        """Get a customer record from the database."""
        return self.Customer.filter_by(customer_id=customer_id).first()

    def export_report(self, case_data):
        """Pass-through to the original API function."""
        return export_report_api(self, case_data)

    def price(self, app_tag, app_tag_version, priority, discount=None):
        """Get the price for a sample from the database."""
        version_obj = (self.ApplicationTagVersion.join(models.ApplicationTagVersion.apptag)
                           .filter(models.ApplicationTagVersion.version == app_tag_version,
                                   models.ApplicationTag.name == app_tag)).first()
        if version_obj:
            log.info("getting price from: %s (%s)", version_obj.apptag.name, version_obj.version)
            full_price = getattr(version_obj, "price_{}".format(priority))
            discount_factor = ((100 - discount) / 100) if discount else 1
            prices = dict(
                kth=full_price * (version_obj.percent_kth / 100) * discount_factor,
                ki=full_price * ((100 - version_obj.percent_kth) / 100) * discount_factor,
                full=full_price * discount_factor,
                discount_factor=discount_factor,
            )
            return prices
        else:
            return None

    def invoice(self, customer_id):
        """Get invoice information about a customer."""
        customer_obj = self.customer(customer_id)
        kth_customer = self.customer('cust999')
        data = dict(
            customer_name=customer_obj.name,
            contact=dict(
                ki=dict(
                    name=customer_obj.invoice_contact.name,
                    email=customer_obj.invoice_contact.email,
                    customer_name=customer_obj.name,
                    reference=customer_obj.invoice_reference,
                    address=customer_obj.invoice_address,
                ),
                kth=dict(
                    name=kth_customer.invoice_contact.name,
                    email=kth_customer.invoice_contact.email,
                    customer_name=kth_customer.name,
                    reference=kth_customer.invoice_reference,
                    address=kth_customer.invoice_address,
                ),
            ),
            agreement=customer_obj.agreement_registration,
        )
        return data

    def add_invoice(self, invoice_data):
        """Create a record for an invoice."""
        hash_factory = hashlib.sha256()
        sample_list = [sample_data['lims_id'] for sample_data in invoice_data['samples']]
        samples_str = ''.join(sample_list)
        hash_factory.update(samples_str.encode())
        invoice_data['invoice_id'] = hash_factory.hexdigest()[:10]

        customer_obj = self.customer(invoice_data['customer_id'])
        new_invoice = models.Invoice(
            customer=customer_obj,
            invoice_id=invoice_data['invoice_id'],
            invoiced_at=datetime.date.today(),
        )
        new_invoice.data = invoice_data
        invoice_obj = self.Invoice.save(new_invoice)
        return invoice_obj


def split_invoice(samples_data, limit=300000):
    """Split invoice based on a cost limit."""
    current_slice = []
    price_sum = 0
    for sample_data in samples_data:
        sample_price = sample_data['prices']['kth'] + sample_data['prices']['ki']

        if price_sum + sample_price > limit:
            yield current_slice
            current_slice = []
            price_sum = 0

        current_slice.append(sample_data)
        price_sum += sample_price

    yield current_slice



# -*- coding: utf-8 -*-
from loqusdb.plugins import MongoAdapter
from loqusdb.utils import load_database
from loqusdb.vcf_tools.vcf import (get_file_handle, check_vcf)


def connect(config):
    """Connect to LoqusDB mongo database."""
    adapter = MongoAdapter()
    adapter.connect(uri=config['loqusdb']['uri'])
    return adapter


def add(adapter, ped_path, vcf_path, case_id):
    """Add observations from a VCF."""
    variant_handle = get_file_handle(vcf_path)
    nr_variants = check_vcf(variant_handle)
    nr_inserted = load_database(
        adapter=adapter,
        variant_file=vcf_path,
        family_file=ped_path,
        family_type='ped',
        gq_treshold=20,
        nr_variants=nr_variants,
        case_id=case_id,
    )

    return dict(variants=nr_variants, inserted=nr_inserted)



# -*- coding: utf-8 -*-
import logging

from clinstatsdb.db import api
from clinstatsdb.analysis.cli import test_analysis
from clinstatsdb.analysis.export import export_run
from clinstatsdb.analysis.mip import process_all
from clinstatsdb.analysis.models import Analysis
import ruamel.yaml

log = logging.getLogger(__name__)


def connect(config):
    """Connect to the database."""
    cgstats_db = api.connect(config['cgstats']['database'])
    return cgstats_db


def add(cgstats_db, case_id, qc_stream, sampleinfo_stream, force=False):
    """Add QC output from an analysis run."""
    sampleinfo = ruamel.yaml.safe_load(sampleinfo_stream)

    if not force and not test_analysis(sampleinfo):
        log.warn("analysis can't be loaded, use '--force'")
    else:
        old_analysis = Analysis.query.filter_by(analysis_id=case_id).first()
        if old_analysis:
            log.warn("removing old analysis")
            old_analysis.delete()
            cgstats_db.commit()

        metrics = ruamel.yaml.safe_load(qc_stream)
        log.debug("parsing analysis output")
        new_analysis = process_all(case_id, sampleinfo, metrics)
        log.info("adding analysis: %s", new_analysis.analysis_id)
        cgstats_db.add_commit(new_analysis)



# -*- coding: utf-8 -*-
import requests


class EMail(object):

    """Interface to Mailgun email API."""

    def __init__(self, config):
        super(EMail, self).__init__()
        self.MAILGUN_API_KEY = config['email']['MAILGUN_API_KEY']
        self.MAILGUN_DOMAIN_NAME = config['email']['MAILGUN_DOMAIN_NAME']

    def send(self, to_addr, subject, text):
        """Send email."""
        url = "https://api.mailgun.net/v3/{}/messages".format(self.MAILGUN_DOMAIN_NAME)
        auth = ('api', self.MAILGUN_API_KEY)
        data = {
            'from': "Mailgun User <mailgun@{}>".format(self.MAILGUN_DOMAIN_NAME),
            'to': to_addr,
            'subject': subject,
            'text': text,
        }
        response = requests.post(url, auth=auth, data=data)
        response.raise_for_status()
        return response

    def deliver(self, ticket_id, family_id):
        """Send delivery email to SupportSystems."""
        to_addr = 'supportsystem@clinicalgenomics.se'
        subject = "[#{}] Delivery to Scout: {}".format(ticket_id, family_id)
        text = """Hello,

        Case {} has been analyzed and uploaded to Scout.

        Your friendly Clinicl Genomics Bot
        """.format(family_id)
        self.send(to_addr, subject, text)



# -*- coding: utf-8 -*-
import logging

from genotype.store import api
from genotype.store.parsemip import parse_mipsex
from genotype.load.vcf import load_vcf

import ruamel.yaml

log = logging.getLogger(__name__)


def connect(config):
    """Connect to the Housekeeper database."""
    genotype_db = api.connect(config['genotype']['database'])
    return genotype_db


def add(genotype_db, raw_bcf_path, force=False):
    """Add genotypes for an analysis run."""
    snps = api.snps()
    analyses = load_vcf(raw_bcf_path, snps)
    for analysis in analyses:
        log.debug('loading VCF genotypes for sample: %s', analysis.sample_id)
        is_saved = api.add_analysis(genotype_db, analysis, replace=force)
        if is_saved:
            log.info('loaded VCF genotypes for sample: %s', analysis.sample_id)
        else:
            log.warn('skipped, found previous analysis: %s', analysis.sample_id)


def add_sex(genotype_db, samples_sex, qcmetrics_stream):
    """Add missing sex information to samples."""
    qcm_data = ruamel.yaml.safe_load(qcmetrics_stream)
    mip_samples_sex = parse_mipsex(qcm_data)
    for sample_id, mip_sex in mip_samples_sex.items():
        genotype_sample = api.sample(sample_id)
        log.info("marking %s as %s", sample_id, samples_sex[sample_id])
        genotype_sample.sex = samples_sex[sample_id]
        genotype_analysis = api.analysis(sample_id, 'sequence').first()
        log.info("marking analysis for %s as %s", sample_id, mip_sex)
        genotype_analysis.sex = mip_sex

    genotype_db.commit()

# {
#     local sample=${1?'missing input - sample name'};
#     local qc_metrics;
#     if qc_metrics=$(housekeeper get --sample "${sample}" --infer-\case --category qc); then
#         local sample_sex=$(cglims get "${sample}" sex);
#         local sequence_sex=$(genotype mip-sex --sample "${sample}" "${qc_metrics}");
#         if [[ -z "${sample_sex}" || -z "${sequence_sex}" ]]; then
#             ( echo "sample sex information missing: ${sample}" 1>&2 );
#             return 1;
#         else
#             genotype add-sex --sample "${sample_sex}" --analysis sequence "${sequence_sex}" "${sample}";
#         fi;
#     else
#         ( echo "can't find QC metrics for: ${sample}" 1>&2 );
#         return 1;
#     fi;
#     return 0
# }



# -*- coding: utf-8 -*-
from sqlalchemy import Column, types, orm, ForeignKey, UniqueConstraint
from sqlservice import declarative_base

Model = declarative_base()


class Case(Model):

    __tablename__ = 'case'

    id = Column(types.Integer, primary_key=True)
    case_id = Column(types.String(256))
    last_analyzed_at = Column(types.DateTime())
    is_prioritized = Column(types.Boolean)
    lims_ok = Column(types.Boolean)
    lims_error = Column(types.Text)

    flowcells = orm.relationship('Flowcell', cascade='all,delete', backref='case')

    def is_ready(self):
        """Check if the case is ready for analysis."""
        linked_fcs = len(self.flowcells) > 0
        flowcells_ok = all(flowcell_obj.has_fastqs for flowcell_obj in self.flowcells)
        return self.lims_ok and linked_fcs and flowcells_ok


class Flowcell(Model):

    __tablename__ = 'flowcell'
    __table_args__ = (
        UniqueConstraint('flowcell_id', 'sample_id', name='_flowcell_sample_uc'),
    )

    id = Column(types.Integer, primary_key=True)
    flowcell_id = Column(types.String(128))
    sample_id = Column(types.String(128))
    case_id = Column(ForeignKey(Case.id))
    is_fetched = Column(types.Boolean, default=False)
    has_fastqs = Column(types.Boolean, default=False)



# -*- coding: utf-8 -*-
import logging

import pymongo
from scout.adapter.mongo import MongoAdapter
from scout.adapter.client import get_connection
from scout.export.panel import export_panels
from scout.load import load_scout

log = logging.getLogger(__name__)


def connect(config, app_key='scout'):
    """Connect to the Scout database."""
    uri = "mongodb://{username}:{password}@{host}:{port}/{database}".format(
        username=config[app_key]['username'],
        password=config[app_key]['password'],
        host=config[app_key]['host'],
        port=config[app_key]['port'],
        database=config[app_key]['database'],
    )
    client = get_connection(uri=uri)
    database = client[config[app_key]['database']]
    mongo_adapter = MongoAdapter(database)
    return mongo_adapter


def get_reruns(adapter):
    """Get cases marked to be rerun."""
    return adapter.case_collection.find({'rerun_requested': True})


def add(scout_db, config_data, threshold=5, force=False):
    """Upload variants for an analysis to the database."""
    config_data['rank_score_threshold'] = threshold
    existing_case = scout_db.case(institute_id=config_data['owner'],
                                  display_name=config_data['family'])
    if existing_case:
        if force or config_data['analysis_date'] > existing_case['analysis_date']:
            log.info("Updating existing Scout case")
            load_scout(scout_db, config_data, update=True)
        else:
            existing_date = existing_case['analysis_date'].date()
            log.warning("Analysis of case already added: %s", existing_date)
    else:
        log.info("Adding new case to Scout")
        load_scout(scout_db, config_data)


def report(scout_db, customer_id, family_id, report_path):
    """Link a delivery report with an existing Scout case."""
    case_obj = scout_db.case(institute_id=customer_id, display_name=family_id)
    if case_obj is None:
        log.error("case not found in database")
        return None

    updated_case = scout_db.case_collection.find_one_and_update(
        {'_id': case_obj['_id']},
        {'$set': {'delivery_report': report_path}},
        return_document=pymongo.ReturnDocument.AFTER
    )
    return updated_case



# -*- coding: utf-8 -*-
import logging

import click
from sqlalchemy.exc import IntegrityError
from sqlalchemy import func
from chanjo.store import api, models
from chanjo.load.sambamba import load_transcripts

from .omim import OMIM_GENE_IDS

log = logging.getLogger(__name__)


class Coverage(api.ChanjoDB):

    def __init__(self, config):
        """Connect to the chanjo coverage database."""
        super(Coverage, self).__init__(uri=config['chanjo']['database'])

    def validate(self, case_id):
        """Validate samples uploaded in the database."""
        query = self.query(
            models.Sample,
            func.avg(models.TranscriptStat.mean_coverage).label('mean_coverage'),
            func.avg(models.TranscriptStat.completeness_10).label('completeness_10'),
        ).join(
            models.TranscriptStat.sample,
            models.TranscriptStat.transcript
        ).filter(
            models.Sample.group_id == case_id,
            models.Transcript.gene_id.in_(OMIM_GENE_IDS)
        ).group_by(models.TranscriptStat.sample_id)
        return query


def connect(config):
    """Connect to the chanjo coverage database."""
    chanjo_db = api.ChanjoDB(uri=config['chanjo']['database'])
    return chanjo_db


def add(chanjo_db, case_id, family_name, sample_id, sample_name, bed_stream, source=None):
    """Add coverage for an analysis to the database."""
    result = load_transcripts(bed_stream, sample_id=sample_id, group_id=case_id, source=source)
    result.sample.name = sample_name
    result.sample.group_name = family_name
    try:
        chanjo_db.add(result.sample)
        with click.progressbar(result.models, length=result.count,
                               label='loading transcripts') as bar:
            for tx_model in bar:
                chanjo_db.add(tx_model)
        chanjo_db.save()
    except IntegrityError as error:
        log.debug(error.message)
        chanjo_db.session.rollback()
        raise error


def sample(sample_id):
    """Get sample from database."""
    return models.Sample.query.get(sample_id)


def delete(chanjo_db, sample_obj):
    """Delete sample from database."""
    chanjo_db.session.delete(sample_obj)
    chanjo_db.save()



# -*- coding: utf-8 -*-
import click

from cg import apps


def parse_caseid(raw_caseid):
    """Parse out parts of the case id."""
    customer_id, raw_familyid = raw_caseid.split('-', 1)
    case_id = raw_caseid.split('--')[0]
    family_parts = raw_familyid.split('--')
    family_id = family_parts[0]
    extra = family_parts[1] if len(family_parts) > 1 else None
    return {
        'raw': {
            'case_id': raw_caseid,
            'family_id': raw_familyid,
        },
        'customer_id': customer_id,
        'family_id': family_id,
        'case_id': case_id,
        'extra': extra,
    }


def check_latest_run(hk_db, context, case_info):
    # get latest analysis for the case
    latest_run = apps.hk.latest_run(hk_db, case_info['raw']['case_id'])
    if latest_run is None:
        click.echo("No run found for the case!")
        context.abort()
    return latest_run



# -*- coding: utf-8 -*-
from collections import Counter, namedtuple
import os.path
from datetime import datetime
import logging

import click
from loqusdb.exceptions import CaseError, VcfError
from housekeeper import exc as hk_exc
from path import Path
import ruamel.yaml

from . import apps
from .utils import parse_caseid, check_latest_run

log = logging.getLogger(__name__)
FakeAnalysis = namedtuple('FakeAnalysis', ['case_id', 'config_path'])


def check_root(context, case_info):
    """Compose and check if root analysis directory for a case exists.

    Aborts the Click CLI context if the family directory doesn't exist.

    Args:
        context (Object): Click context object
        case_info (dict): Parsed case id information

    Returns:
        path: root analysis directory for the family/case
    """
    root_dir = Path(context.obj['analysis_root'])
    family_dir = root_dir.joinpath(case_info['customer_id'], case_info['raw']['family_id'])
    if not family_dir.exists():
        log.error("family directory not found: %s", family_dir)
        context.abort()
    return family_dir


@click.command('mip-config')
@click.option('-p', '--print', 'print_output', is_flag=True, help='print to console')
@click.argument('case_id')
@click.pass_context
def mip_config(context, print_output, case_id):
    """Generate a MIP config from LIMS data."""
    case_info = parse_caseid(case_id)

    log.debug("get config data: %s", case_info['case_id'])
    lims_api = apps.lims.connect(context.obj)
    data = apps.lims.config(lims_api, case_info['customer_id'], case_info['family_id'])

    if case_info['extra']:
        log.info("update config with suffix: %s", case_info['extra'])
        apps.lims.extend_case(data, case_info['extra'])

    raw_output = ruamel.yaml.round_trip_dump(data, indent=4, block_seq_indent=2)
    if print_output:
        click.echo(raw_output.decode())
    else:
        family_dir = check_root(context, case_info)
        config_file = "{}_pedigree.yaml".format(case_info['raw']['family_id'])
        config_path = family_dir.joinpath(config_file)
        with config_path.open('w') as out_handle:
            click.echo(raw_output.decode(), file=out_handle)
        log.info("wrote config to: %s", config_path)


@click.command()
@click.option('--source', '-s', type=click.Choice(['prod', 'archive']), default='prod')
@click.pass_context
def reruns(context, source='prod'):
    """Return reruns marked in Scout (old)."""
    scout_db = apps.scoutprod.connect(context.obj)
    if source == 'prod':
        for scout_case in apps.scoutprod.get_reruns(scout_db):
            click.echo(scout_case['_id'])

    elif source == 'archive':
        scoutold_db = apps.scoutold.connect(context.obj)
        for scout_case in apps.scoutold.get_reruns(scoutold_db):
            case_id = scout_case['_id'].replace('_', '-', 1)
            # lookup requested case in current Scout
            if scout_db.case(case_id=scout_case['_id']):
                pass
            else:
                click.echo(case_id)


@click.command('mip-panel')
@click.option('-p', '--print', 'print_output', is_flag=True, help='print to console')
@click.argument('case_id')
@click.pass_context
def mip_panel(context, print_output, case_id):
    """Generate an aggregated panel for MIP."""
    case_info = parse_caseid(case_id)
    lims_api = apps.lims.connect(context.obj)
    samples = lims_api.case(case_info['customer_id'], case_info['raw']['family_id'])

    # fetch default panels
    default_panels = set()
    for lims_sample in samples:
        default_panels.update(apps.lims.sample_panels(lims_sample))

    # convert default panels to all panels
    all_panels = apps.lims.convert_panels(case_info['customer_id'], default_panels)
    log.debug("determined panels to use: %s", ', '.join(all_panels))

    adapter = apps.scoutprod.connect(context.obj)
    bed_lines = apps.scoutprod.export_panels(adapter, all_panels)

    if print_output:
        for bed_line in bed_lines:
            click.echo(bed_line)
    else:
        family_dir = check_root(context, case_info)
        panel_path = family_dir.joinpath('aggregated_master.bed')
        with panel_path.open('w') as out_handle:
            click.echo('\n'.join(bed_lines).decode(), file=out_handle)
        log.info("wrote aggregated gene panel: %s", panel_path)


@click.command()
@click.option('-a', '--answered-out', is_flag=True, help='fill in answered out status')
@click.argument('case_id')
@click.pass_context
def update(context, answered_out, case_id):
    """Fill in information in Housekeeper."""
    if answered_out:
        hk_db = apps.hk.connect(context.obj)
        lims_api = apps.lims.connect(context.obj)
        log.debug("get case from housekeeper")
        hk_case = apps.hk.api.case(case_id)
        log.debug("loop over related samples from most recent run")
        delivery_dates = []
        hk_run = hk_case.current
        for hk_sample in hk_run.samples:
            log.debug("lookup if sample has been delivered in LIMS")
            delivery_date = lims_api.is_delivered(hk_sample.lims_id)
            if delivery_date is None:
                log.warn("sample not delivered: %s", hk_sample.lims_id)
                context.abort()
            delivery_dates.append(delivery_date)
        latest_date = sorted(delivery_dates)[-1]
        log.debug("fillin answered out date in HK")
        hk_run.answeredout_at = datetime.combine(latest_date, datetime.min.time())
        hk_db.commit()
        log.info("run 'answered out' date updated: %s", case_id)


@click.command()
@click.argument('process_id')
@click.pass_context
def check(context, process_id):
    """Check samples in LIMS and optionally update them."""
    admin_db = apps.admin.Application(context.obj)
    lims_api = apps.lims.connect(context.obj)
    samples = list(apps.lims.process_to_samples(lims_api, process_id))
    uniq_tags = set(sample['sample'].udf['Sequencing Analysis'] for sample in samples)
    apptag_map = admin_db.map_apptags(uniq_tags)
    apps.lims.check_samples(lims_api, samples, apptag_map)


@click.command()
@click.option('-s', '--setup/--no-setup', default=True,
              help='perform setup before starting analysis')
@click.option('--execute/--no-execute', default=True, help='skip running MIP')
@click.option('-f', '--force', is_flag=True, help='skip pre-analysis checks')
@click.option('--hg38', is_flag=True, help='run with hg38 settings')
@click.option('-e', '--email', help='email to send errors to')
@click.argument('case_id')
@click.pass_context
def start(context, setup, execute, force, hg38, email, case_id):
    """Start a MIP analysis."""
    case_info = parse_caseid(case_id)
    if setup:
        log.info('generate aggregated gene panel')
        context.invoke(mip_panel, case_id=case_id)
        log.info('generate analysis pedigree')
        context.invoke(mip_config, case_id=case_id)

    lims_api = apps.lims.connect(context.obj)
    lims_data = apps.lims.start(lims_api, case_info['customer_id'], case_info['raw']['family_id'])

    log.info("start analysis for: %s", case_id)
    apps.tb.start_analysis(context.obj, case_info, hg38=hg38, force=force, execute=execute,
                           skip_validation=lims_data['is_external'],
                           prioritize=lims_data['is_prio'], email=email)


@click.command('auto-start')
@click.option('--dry-run', is_flag=True, help='skip launching new runs')
@click.option('-e', '--email', help='email to send errors to')
@click.option('-f', '--force', is_flag=True, help='start all available runs')
@click.option('-r', '--running', default=20, help='how many runs to start at one time')
@click.pass_context
def auto_start(context, dry_run, email, force, running):
    """Automatically start analysis for sequenced cases."""
    hk_db = apps.hk.connect(context.obj)
    tb_db = apps.tb.connect(context.obj)
    cases = apps.hk.to_analyze(hk_db)

    log.info("%s cases can be started", cases.count())
    for case_obj in cases:
        analyses_running = apps.tb.analyses_running(tb_db)
        if not force and analyses_running >= running:
            log.warn("already %s analyses running, pausing", analyses_running)
            break

        log.debug("working on case: %s", case_obj.name)
        if not all(sample_obj.sequenced_at for sample_obj in case_obj.samples):
            log.warn("all samples not sequenced: %s", case_obj.name)
        elif len(case_obj.runs) != 0:
            log.warn("case already analyzed: %s", case_obj.name)
        elif apps.tb.api.is_running(case_obj.name):
            log.debug("already running, skipping: %s", case_obj.name)
        elif apps.tb.latest_status(tb_db, case_obj.name, 'failed'):
            log.debug("latest run has failed, skipping: %s", case_obj.name)
        elif apps.tb.latest_status(tb_db, case_obj.name, 'completed'):
            log.debug("latest run completed, skipping: %s", case_obj.name)
        else:
            log.info("starting case: %s", case_obj.name)
            if not dry_run:
                try:
                    context.invoke(start, case_id=case_obj.name, email=email)
                except ValueError as error:
                    log.error(error.args[0])
                except KeyError as error:
                    log.error("error creating config, sample: {}".format(error.args[0]))


@click.command()
@click.argument('mip_config_path', type=click.Path(exists=True), required=False)
@click.pass_context
def keep(context, mip_config_path):
    """Add recently completed analyses to Housekeeper."""
    tb_db = apps.tb.connect(context.obj)
    hk_db = apps.hk.connect(context.obj)
    if mip_config_path:
        analyses = [FakeAnalysis(os.path.basename(mip_config_path), mip_config_path)]
    else:
        analyses = apps.tb.recently_completed(tb_db)
    for analysis_obj in analyses:
        log.debug("working on case: %s", analysis_obj.case_id)
        try:
            root_path = context.obj['housekeeper']['root']
            with open(analysis_obj.config_path, 'r') as in_handle:
                mip_config_data = ruamel.yaml.safe_load(in_handle)
            apps.hk.add(hk_db, root_path, mip_config_data)
            log.info("added case to housekeeper: %s", analysis_obj.case_id)
        except hk_exc.AnalysisConflictError as error:
            log.debug('analysis already added to housekeeper')
        except hk_exc.MalformattedPedigreeError as error:
            log.error("bad PED formatting: %s", error.message)
        except hk_exc.AnalysisNotFinishedError as error:
            log.error("analysis not finished: %s", error.message)


@click.command()
@click.option('-f', '--force', is_flag=True, help='skip pre-upload checks')
@click.argument('case_id')
@click.pass_context
def coverage(context, force, case_id):
    """Upload coverage for an analysis (latest)."""
    chanjo_db = apps.coverage.connect(context.obj)
    hk_db = apps.hk.connect(context.obj)
    lims_api = apps.lims.connect(context.obj)
    case_info = parse_caseid(case_id)
    latest_run = check_latest_run(hk_db, context, case_info)

    coverage_date = latest_run.extra.coverage_date
    if not force and coverage_date:
        click.echo("Coverage already added for run: {}".format(coverage_date.date()))
    else:
        for sample_data in apps.hk.coverage(hk_db, latest_run):
            chanjo_sample = apps.coverage.sample(sample_data['sample_id'])
            if chanjo_sample:
                log.warn("removing existing sample: %s", chanjo_sample.id)
                apps.coverage.delete(chanjo_db, chanjo_sample)

            lims_sample = lims_api.sample(sample_data['sample_id'])
            log.info("adding coverage for sample: %s", sample_data['sample_id'])
            with open(sample_data['bed_path']) as bed_stream:
                apps.coverage.add(
                    chanjo_db,
                    case_id=case_info['raw']['case_id'],
                    family_name=case_info['raw']['family_id'],
                    sample_id=sample_data['sample_id'],
                    sample_name=lims_sample.name,
                    bed_stream=bed_stream,
                    source=sample_data['bed_path']
                )

        log.info("marking coverage added for case: %s", case_info['raw']['case_id'])
        latest_run.extra.coverage_date = latest_run.analyzed_at
        hk_db.commit()


@click.command()
@click.option('-f', '--force', is_flag=True, help='skip pre-upload checks')
@click.argument('case_id')
@click.pass_context
def genotypes(context, force, case_id):
    """Add VCF genotypes for an analysis run (latest)."""
    genotype_db = apps.gt.connect(context.obj)
    hk_db = apps.hk.connect(context.obj)
    lims_api = apps.lims.connect(context.obj)
    case_info = parse_caseid(case_id)
    latest_run = check_latest_run(hk_db, context, case_info)

    if not force and latest_run.extra.genotype_date:
        click.echo("Genotypes already added for run: {}"
                   .format(latest_run.extra.genotype_date.date()))
    else:
        assets = apps.hk.genotypes(hk_db, latest_run)
        apps.gt.add(genotype_db, assets['bcf_path'], force=force)
        samples_sex = apps.lims.case_sexes(lims_api, case_info['customer_id'],
                                           case_info['family_id'])
        with open(assets['qc_path']) as qcmetrics_stream:
            apps.gt.add_sex(genotype_db, samples_sex, qcmetrics_stream)

        log.info("marking genotypes added for case: %s", case_info['raw']['case_id'])
        latest_run.extra.genotype_date = latest_run.analyzed_at
        hk_db.commit()


@click.command()
@click.option('-f', '--force', is_flag=True, help='skip pre-upload checks')
@click.argument('case_id')
@click.pass_context
def qc(context, force, case_id):
    """Add QC metrics for an anlyais run (latest)."""
    hk_db = apps.hk.connect(context.obj)
    cgstats_db = apps.qc.connect(context.obj)
    case_info = parse_caseid(case_id)
    latest_run = check_latest_run(hk_db, context, case_info)

    if not force and latest_run.extra.qc_date:
        click.echo("QC already added for run: {}".format(latest_run.extra.qc_date.date()))
    else:
        assets = apps.hk.qc(hk_db, latest_run)
        with open(assets['qc_path']) as qc_stream, open(assets['sampleinfo_path']) as si_stream:
            apps.qc.add(cgstats_db, case_info['raw']['case_id'], qc_stream, si_stream, force=force)

        log.info("marking qc added for case: %s", case_info['raw']['case_id'])
        latest_run.extra.qc_date = latest_run.analyzed_at
        hk_db.commit()


@click.command()
@click.option('-f', '--force', is_flag=True, help='skip pre-upload checks')
@click.option('-t', '--threshold', default=5, help='rank score threshold')
@click.argument('case_id')
@click.pass_context
def visualize(context, force, threshold, case_id):
    """Add data to Scout for an analysis run (latest)."""
    hk_db = apps.hk.connect(context.obj)
    scout_db = apps.scoutprod.connect(context.obj)
    case_info = parse_caseid(case_id)
    latest_run = check_latest_run(hk_db, context, case_info)

    if not latest_run.pipeline_version.startswith('v4'):
        log.error("unsupported pipeline version: %s", latest_run.pipeline_version)
        context.abort()

    if not force and latest_run.extra.visualizer_date:
        click.echo("Run already added to scout: {}"
                   .format(latest_run.extra.visualizer_date.date()))
    else:
        config_path = apps.hk.visualize(hk_db, latest_run,
                                        context.obj['housekeeper']['madeline_exe'],
                                        context.obj['housekeeper']['root'])
        with open(config_path) as config_stream:
            config_data = ruamel.yaml.safe_load(config_stream)

        config_data['rank_score_threshold'] = threshold
        apps.scoutprod.add(scout_db, config_data, force=force)

        log.info("marking visualize added for case: %s", case_info['raw']['case_id'])
        latest_run.extra.visualizer_date = latest_run.analyzed_at
        hk_db.commit()


@click.command('delivery-report')
@click.argument('case_id')
@click.pass_context
def delivery_report(context, case_id):
    """Generate delivery report for the latest analysis."""
    log.info('connecting to databases')
    hk_db = apps.hk.connect(context.obj)
    lims_api = apps.lims.connect(context.obj)
    cgstats_db = apps.qc.connect(context.obj)
    admin_db = apps.admin.Application(context.obj)
    scout_db = apps.scoutprod.connect(context.obj)
    case_info = parse_caseid(case_id)
    latest_run = check_latest_run(hk_db, context, case_info)

    log.info('fetching data from LIMS')
    case_data = apps.lims.export(lims_api, case_info['customer_id'], case_info['family_id'])
    log.info('fetching data from cgstats')
    case_data = apps.qc.export_run(cgstats_db, case_data)
    log.info('generating report in cgadmin')
    template_out = admin_db.export_report(case_data)

    run_root = apps.hk.rundir(context.obj, latest_run)
    report_file = os.path.join(run_root, 'delivery-report.html')
    log.info("saving report to: %s", report_file)
    with click.open_file(report_file, mode='w', encoding='utf-8') as out_handle:
        out_handle.write(template_out)

    log.info("adding report to housekeeper")
    apps.hk.add_asset(hk_db, latest_run, report_file, 'export', archive_type='result')
    apps.scoutprod.report(scout_db, case_info['customer_id'], case_info['family_id'], report_file)


@click.command()
@click.argument('case_id')
@click.pass_context
def observations(context, case_id):
    """Add observations for a case."""
    case_info = parse_caseid(case_id)
    hk_db = apps.hk.connect(context.obj)
    latest_run = check_latest_run(hk_db, context, case_info)
    paths = apps.hk.observations(hk_db, latest_run)
    loqus_db = apps.loqus.connect(context.obj)

    existing_case = loqus_db.case(dict(case_id=case_id))
    if existing_case:
        log.warn("found observations for existing case - skipping: %s", case_id)
    else:
        try:
            apps.loqus.add(loqus_db, paths['ped'], paths['vcf'], case_id=case_id)
        except (CaseError, VcfError) as error:
            log.error(error)
            context.abort()


@click.command()
@click.option('-f', '--force', is_flag=True, help='skip pre-upload checks')
@click.argument('case_id')
@click.pass_context
def add(context, force, case_id):
    """Uplaod analysis results (latest) to various databases."""
    hk_db = apps.hk.connect(context.obj)
    admin_db = apps.admin.Application(context.obj)
    tb_db = apps.tb.connect(context.obj)
    case_info = parse_caseid(case_id)
    latest_run = check_latest_run(hk_db, context, case_info)

    if not force and latest_run.delivered_at:
        log.warn("Run already delivered: %s", latest_run.delivered_at.date())
    else:
        log.info("Add coverage")
        context.invoke(coverage, force=force, case_id=case_id)
        log.info("Add genotypes and sex information")
        context.invoke(genotypes, force=force, case_id=case_id)
        log.info("Add QC metrics")
        context.invoke(qc, force=force, case_id=case_id)
        if admin_db.customer(case_info['customer_id']).scout_access:
            if not force:
                log.info("Validate quality criteria")
                context.invoke(validate, case_id=case_id)
            else:
                log.warn("Skipping quality validation!")

            log.info("Add case and variants to Scout")
            context.invoke(visualize, force=force, case_id=case_id)
            log.info("Add delivery report to Scout upload")
            context.invoke(delivery_report, case_id=case_id)

            log.info("Add observations to local database, loqusdb")
            context.invoke(observations, case_id=case_id)

            # log.info("Send email about successful delivery")
            # email = apps.email.EMail(context.obj)
            # lims_api = apps.lims.connect(context.obj)
            # lims_samples = lims_api.case(case_info['customer_id'], case_info['raw']['family_id'])
            # ticket_id = lims_samples[0].project.name
            # email.deliver(ticket_id, case_info['raw']['family_id'])

        log.info("marking analysis run as delivered: %s", case_info['raw']['case_id'])
        latest_run.delivered_at = datetime.now()
        hk_db.commit()

        log.info("commenting on run log in trailblazer")
        analysis_log = (apps.tb.api.Analysis.query
                            .filter_by(case_id=case_info['raw']['case_id'],
                                       started_at=latest_run.analyzed_at)
                            .first())
        if analysis_log:
            message = 'analysis delivered! /cg'
            if analysis_log.comment:
                analysis_log.comment = "{}\n\n{}".format(message, analysis_log.comment)
            else:
                analysis_log.comment = message
            tb_db.commit()


@click.command()
@click.argument('case_id')
@click.pass_context
def validate(context, case_id):
    """Validate samples in a case."""
    case_info = parse_caseid(case_id)
    lims_api = apps.lims.connect(context.obj)
    lims_data = apps.lims.validate(lims_api, case_info['customer_id'],
                                   case_info['raw']['family_id'])

    chanjo_db = apps.coverage.Coverage(context.obj)
    coverage_query = chanjo_db.validate(case_id)

    results = Counter()
    for data in coverage_query:
        sample_id = data.Sample.id
        log.info("validating sample: %s", sample_id)
        if lims_data[sample_id]['is_external']:
            log.info("external sample, skipping quality checks")
            results.update(['SKIP'])
            continue
        elif lims_data[sample_id]['is_panel']:
            log.debug("TGA sample")
            if lims_data[sample_id]['reads'] >= 75e6 and data.completeness_10 < 0.96:
                log.error("completeness too low: %s < 0.96", data.completeness_10)
                results.update(['FAIL'])
            elif lims_data[sample_id]['reads'] >= 30e6 and data.completeness_10 < 0.95:
                log.error("completeness too low: %s < 0.95", data.completeness_10)
                results.update(['FAIL'])
            elif data.completeness_10 < 0.8:
                log.error("completeness too low: %s < 0.8", data.completeness_10)
                results.update(['FAIL'])
            else:
                log.info('coverage OK!')
                results.update(['PASS'])
        else:
            log.debug('WGS sample')
            if lims_data[sample_id]['is_low_input'] and data.mean_coverage < 22.5:
                log.debug('low input sample')
                log.error("mean coverage too low: %s < 22.5", data.mean_coverage)
                results.update(['FAIL'])
            elif data.mean_coverage < 26:
                log.error("mean coverage too low: %s < 26", data.mean_coverage)
                results.update(['FAIL'])
            else:
                log.info('coverage OK!')
                results.update(['PASS'])

    samples_no = len(lims_data)
    skipped_no = results['SKIP'] or 0
    if skipped_no > 0:
        log.warn("skipped %s/%s samples", results['SKIP'] or 0, samples_no)

    log.info("passed %s/%s samples", results['PASS'] or 0, samples_no)

    failed_no = results['FAIL'] or 0
    if failed_no > 0:
        log.error("failed %s/%s samples", results['FAIL'] or 0, samples_no)
        context.abort()



# -*- coding: utf-8 -*-
import datetime
import logging

from trailblazer.store import api
from trailblazer.add.commit import commit_analysis
from trailblazer.analyze.cli import environ_email
from trailblazer.analyze.start import start_mip, build_pending
from path import Path
import ruamel.yaml

log = logging.getLogger(__name__)


def connect(config):
    """Connect to the Trailblazer database."""
    tb_db = api.connect(config['trailblazer']['database'])
    return tb_db


def start_analysis(config, case_info, hg38=False, execute=True, force=False,
                   skip_validation=False, prioritize=False, email=None):
    """Start analysis run."""
    tb_db = connect(config)
    customer_id = case_info['customer_id']
    family_id = case_info['raw']['family_id']
    case_id = case_info['raw']['case_id']
    cc_path = Path(config['trailblazer']['analysis_root']).joinpath(customer_id)
    if not force:
        check_setup(tb_db, case_id)

    pedigree_path = cc_path.joinpath("{0}/{0}_pedigree.yaml".format(family_id))
    if not pedigree_path.exists():
        raise ValueError("pedigree YAML doesn't exist: {}".format(pedigree_path))
    with pedigree_path.open() as in_handle:
        ped_data = ruamel.yaml.round_trip_load(in_handle)
        analysis_type = case_analysis_type(ped_data)
        # set max guassian flag for SNVs if WGS analysis
        max_gaussian = True if analysis_type == 'wgs' else False

    global_config = (config['trailblazer']['mip_config_hg38'] if hg38 else
                     config['trailblazer']['mip_config'])
    executable = config['trailblazer']['mip_exe']
    email = email or environ_email()

    flags = [('--qccollect_skip_evaluation', '1')] if skip_validation else None
    priority = 'high' if prioritize else 'normal'
    process = start_mip(config=global_config, family_id=family_id, ccp=cc_path,
                        executable=executable, email=email, max_gaussian=max_gaussian,
                        execute=execute, flags=flags, priority=priority)

    if execute:
        process.wait()
        if process.returncode != 0:
            raise ValueError("check output, error starting analysis: {}"
                             .format(case_id))

        # add pending entry to database
        new_entry = build_pending(case_id, cc_path)
        if email:
            user = api.user(email)
            new_entry.user = user

        commit_analysis(tb_db, new_entry)
        tb_db.commit()


def check_setup(tb_db, case_id):
    """Perform pre-analysis checks."""
    if api.is_running(case_id):
        raise ValueError("case already running: {}".format(case_id))


def case_analysis_type(ped_data):
    """Get overall analysis type for a case."""
    analysis_types = set(sample['analysis_type'] for sample in ped_data['samples'])
    if len(analysis_types) == 1:
        return analysis_types.pop()
    elif 'wgs' in analysis_types:
        return 'wgs'
    else:
        raise ValueError("unknown analysis type: {}".format(analysis_types))


def recently_completed(tb_db):
    """Return recently completed analyses."""
    today = datetime.date.today()
    week_ago = today - datetime.timedelta(days=3)
    analysis_q = api.analyses(since=week_ago, is_ready=True)
    return analysis_q


def analyses_running(tb_db):
    running_q = api.analyses(status=['pending', 'running'])
    return running_q.count()


def latest_status(tb_db, case_name, status):
    """Check if the latest run of a case has failed."""
    latest_case_run = api.analyses(analysis_id=case_name).first()
    if latest_case_run is None:
        return False
    elif latest_case_run.status == status:
        return True
    else:
        return False



# -*- coding: utf-8 -*-
import logging
import pkg_resources

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from housekeeper.pipelines.mip4.scout import prepare_scout
from housekeeper.pipelines.cli import LOADERS, link_records
from housekeeper.pipelines.general import commit_analysis
import ruamel.yaml

from cg.exc import MissingFileError

log = logging.getLogger(__name__)


def connect(config):
    """Connect to the Housekeeper database."""
    hk_db = api.manager(config['housekeeper']['database'])
    return hk_db


def latest_run(hk_db, case_id):
    """Get the latest analysis for a case."""
    case_obj = api.case(case_id)
    if case_obj is None or len(case_obj.runs) == 0:
        return None
    return case_obj.runs[0]


def coverage(hk_db, analysis_obj):
    """Parse analysis record for uploading coverage."""
    for sample_obj in analysis_obj.samples:
        coverage_bed = api.assets(sample=sample_obj.lims_id, category='coverage',
                                  run_id=analysis_obj.id).first()
        if coverage_bed is None:
            log.warn("no coverage output found for sample: %s", sample_obj.lims_id)
            continue

        yield dict(sample_id=sample_obj.lims_id, bed_path=coverage_bed.path)


def genotypes(hk_db, analysis_obj):
    """Parse analysis record for uploading genotypes."""
    # first try to find a "gBCF"
    bcf_obj = api.assets(run_id=analysis_obj.id, category='gbcf').first()
    if bcf_obj is None:
        log.warn("can't find gBCF, looking up raw BCF file")
        bcf_obj = api.assets(run_id=analysis_obj.id, category='bcf-raw').first()

    if bcf_obj is None:
        log.error("BCF file missing")
        raise MissingFileError("gBCF/BCF not found")

    qc_obj = api.assets(run_id=analysis_obj.id, category='qc').first()
    return dict(bcf_path=bcf_obj.path, qc_path=qc_obj.path)


def qc(hk_db, analysis_obj):
    """Parse analysis record for adding QC output."""
    qc_obj = api.assets(run_id=analysis_obj.id, category='qc').first()
    sampleinfo_obj = api.assets(run_id=analysis_obj.id, category='sampleinfo').first()
    return dict(qc_path=qc_obj.path, sampleinfo_path=sampleinfo_obj.path)


def visualize(hk_db, analysis_obj, madeline_exe, root_path):
    """Parse analysis record for uploading to Scout."""
    log.info("create/replace Scout load config")
    existing_conf = api.assets(run_id=analysis_obj.id, category='scout-config').first()
    if existing_conf:
        log.info("delete existing scout config: %s", existing_conf.path)
        api.delete_asset(existing_conf)
    existing_mad = api.assets(run_id=analysis_obj.id, category='madeline').first()
    if existing_mad:
        log.info("delete existing madeline: %s", existing_mad.path)
        api.delete_asset(existing_mad)
    hk_db.commit()

    prepare_scout(analysis_obj, root_path, madeline_exe)
    hk_db.commit()

    scout_config = api.assets(run_id=analysis_obj.id, category='scout-config').first()
    return scout_config.path


def observations(hk_db, analysis_obj):
    """Parse analysis record for uploading to LoqusDB."""
    existing_ped = api.assets(run_id=analysis_obj.id, category='pedigree').first()
    existing_vcf = api.assets(run_id=analysis_obj.id, category='vcf-research-bin').first()
    return dict(ped=existing_ped.path, vcf=existing_vcf.path)


def rundir(config, analysis_obj):
    """Get root directory for a run."""
    root_path = get_rundir(config['housekeeper']['root'], analysis_obj.case.name, analysis_obj)
    return root_path


def add_asset(hk_db, analysis_obj, asset_path, category, archive_type=None, sample=None):
    """Add a new asset to an existing analysis run."""
    # check existing asset
    existing_asset = api.assets(run_id=analysis_obj.id, category=category).first()
    if existing_asset:
        log.warn("existing asset exists!")
        return existing_asset

    new_asset = api.add_asset(analysis_obj, asset_path, category, archive_type, sample=sample)
    new_asset.path = asset_path
    analysis_obj.assets.append(new_asset)
    log.info("add asset: %s", new_asset.path)
    hk_db.commit()
    return new_asset


def to_analyze(hk_db):
    """Fetch cases to be analyzed."""
    cases_q = api.cases(missing='analyzed', onhold=False)
    return cases_q


def add(hk_db, root_path, analysis_config, pipeline='mip4', force=False):
    """Add analyses from different pipelines."""
    default_ref = "pipelines/references/{}.yaml".format(pipeline)
    references = pkg_resources.resource_string('housekeeper', default_ref)
    reference_data = ruamel.yaml.safe_load(references)
    loader = LOADERS[pipeline]
    data = loader(analysis_config, reference_data, force=force)
    records = link_records(data)
    case_name = records['case'].name
    commit_analysis(hk_db, root_path, records['case'], records['run'])
    log.info("added new analysis: %s", case_name)
    sample_ids = ', '.join(sample.lims_id for sample in records['run'].samples)
    log.info("including samples: %s", sample_ids)



# -*- coding: utf-8 -*-
from sqlservice import SQLClient

from .models import Model, Case


class QueueApi(object):

    """docstring for QueueApi"""

    def __init__(self, db_uri):
        super(QueueApi, self).__init__()
        self.db_uri = db_uri
        self.db = None
        self.connect()

    def connect(self):
        """Connect to the database."""
        self.db = SQLClient({'SQL_DATABASE_URI': self.db_uri}, model_class=Model)

    def case(self, case_id):
        """Fetch a case from the database."""
        return self.db.Case.get(case_id)

    def case_by_name(self, case_name):
        """Fetch a case by name."""
        return self.db.Case.filter_by(case_id=case_name).first()

    def save_case(self, case_data):
        """Save a new case to the database."""
        return self.db.Case.save(case_data)

    def cases(self, lims_ok=None):
        """Return cases in order of highest priority."""
        query = self.db.Case.order_by(Case.is_prioritized.desc(),
                                      Case.last_analyzed_at.desc())
        if lims_ok is not None:
            query = query.filter_by(lims_ok=lims_ok)
        return query

    def flowcell(self, flowcell_id, sample_id):
        """Return result for a flowcell and sample."""
        return self.db.Flowcell.filter_by(flowcell_id=flowcell_id, sample_id=sample_id).first()

    def save_flowcell(self, flowcell_data):
        """Save a new flowcell (+ sample) entry to the database."""
        return self.db.Flowcell.save(flowcell_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, 'cg', 'version.py')) as in_handle:
    exec(in_handle.read(), about)

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist bdist_wheel upload")
    sys.exit()


setup(
    name='cg',
    version=about['__version__'],
    description='Glue CLI for Clinical Genomics apps.',
    author='Robin Andeer',
    author_email='robin.andeer@scilifelab.se',
    url='https://github.com/Clinical-Genomics/cg',
    include_package_data=True,
    zip_safe=False,
    packages=find_packages(exclude=('tests*', 'docs', 'examples')),
    entry_points={
        'console_scripts': ['cg=cg:cli'],
    },
    install_requires=[
        'click>=6.7',
        'click_completion',
        'crayons',
        'ruamel.yaml',
        'pymongo',
        'coloredlogs',
        'trailblazer>=1.4.0',
        'scout-browser',
        'pymysql',
        'loqusdb>=0.5',
        'housekeeper>=1.2.0',
        'path.py',
        'cglims>=1.2.0',
        'sqlservice',
        'cgadmin>=1.0.0',
        'chanjo>=4.1.1',
        'requests',
        'genotype>=2.1.0',
        'genologics',
        'scout>=3.0.7',
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
)



# -*- coding: utf-8 -*-
from __future__ import division
from datetime import date
import logging

import click
from cglims.exc import MissingLimsDataException

from cg import apps
from cg.exc import MultipleCustomersError

log = logging.getLogger(__name__)
COSTCENTERS = ['kth', 'ki']


@click.command()
@click.argument('process_id')
@click.pass_context
def invoice(context, process_id):
    """Generate invoices from a LIMS process."""
    log.debug("connecting to databases")
    lims_api = apps.lims.connect(context.obj)
    admin_db = apps.admin.Application(context.obj)

    log.info("getting invoice data from process")
    lims_data = apps.lims.invoice_process(lims_api, process_id)

    log.info("generate invoice data from LIMS and cgadmin")
    lims_samples = lims_data['lims_samples']
    try:
        invoices = generate_invoices(lims_api, admin_db, lims_samples,
                                     discount=lims_data['discount'])
    except (MultipleCustomersError, MissingLimsDataException) as error:
        log.error(error.message)
        context.abort()

    invoice_ids = []
    for invoice_data in invoices:
        new_invoice = admin_db.add_invoice(invoice_data)
        log.info("prepared new invoice: %s", new_invoice.invoice_id)
        invoice_ids.append(new_invoice.invoice_id)

        # for sample_data in invoice_data['samples']:
        #     lims_sample = lims_api.sample(sample_data['lims_id'])
        #     lims_sample.udf['Invoice reference'] = new_invoice.invoice_id
        #     lims_sample.put()

    lims_data['lims_process'].udf['Invoice reference'] = ', '.join(invoice_ids)
    lims_data['lims_process'].udf['Invoicing date'] = date.today()
    lims_data['lims_process'].put()


def generate_invoices(lims_api, admin_db, lims_samples, discount=None):
    """Generate an invoice from a list of samples."""
    lims_data = apps.lims.invoice(lims_samples)
    base_invoice_data = admin_db.invoice(lims_data['customer_id'])
    base_invoice_data['customer_id'] = lims_data['customer_id']

    for data in lims_data['samples']:
        log.debug("including sample: %s", data['lims_id'])
        data['prices'] = admin_db.price(data['application_tag'], data['application_tag_version'],
                                        data['priority'], discount=discount)

    limit = 300000 if lims_data['customer_id'] == 'cust002' else 1000000000  # inf
    for samples_data in apps.admin.split_invoice(lims_data['samples'], limit=limit):
        invoice_data = base_invoice_data.copy()
        invoice_data['samples'] = samples_data
        yield invoice_data



# -*- coding: utf-8 -*-
import logging

from dateutil.parser import parse as parse_date
from cglims import api
from cglims.config import basic_config
from cglims.exc import MissingLimsDataException
from cglims.export import export_case
from cglims.panels import convert_panels
from cglims.check import check_sample, process_samples
from genologics.entities import Process

from cg.exc import MultipleCustomersError

log = logging.getLogger(__name__)
config = basic_config


def connect(config):
    """Connect to LIMS."""
    return api.connect(config['cglims']['genologics'])


def case_sexes(lims_api, customer_id, family_name):
    """Fetch sample sex information from LIMS."""
    sample_sex = {}
    for lims_sample in lims_api.case(customer_id, family_name):
        sample_data = api.ClinicalSample(lims_sample).to_dict()
        sample_sex[sample_data['sample_id']] = sample_data['sex']
    return sample_sex


def extend_case(config_data, suffix):
    """Modify ids for downstream extensions."""
    config_data['family'] = "{}--{}".format(config_data['family'], suffix)
    for sample_data in config_data['samples']:
        sample_data['sample_id'] = "{}--{}".format(sample_data['sample_id'], suffix)


def sample_panels(lims_sample):
    """Get gene panels from a sample."""
    raw_panels = lims_sample.udf.get('Gene List')
    default_panels = raw_panels.split(';') if raw_panels else []
    additional_panel = lims_sample.udf.get('Additional Gene List')
    if additional_panel:
        default_panels.append(additional_panel)
    return default_panels


def check_samples(lims_api, samples, apptag_map):
    """Check if new samples in LIMS are correct."""
    for sample_dict in samples:
        lims_sample = sample_dict['sample']
        apptag_version = apptag_map[lims_sample.udf['Sequencing Analysis']]
        check_sample(lims_api, lims_sample, lims_artifact=sample_dict.get('artifact'),
                     update=True, version=apptag_version)


def process_to_samples(lims_api, process_id):
    """Get all LIMS samples and artifacts from a process."""
    lims_process = Process(lims_api, id=process_id)
    return process_samples(lims_process)


def check_config(lims_api, customer_id, family_id):
    """Check generation of pedigree config."""
    try:
        config_data = config(lims_api, customer_id, family_id)
    except MissingLimsDataException as error:
        log.error(error.args[0])
        return {'success': False, 'message': error.args[0]}
    except KeyError as error:
        log.error("UDF - %s", error.args[0])
        return {'success': False, 'message': error.args[0]}
    except AssertionError as error:
        if error.args[0].endswith('set()'):
            log.error('missing customer UDF')
            return {'success': False, 'message': 'missing customer UDF'}
        else:
            log.error(error.args[0])
            return {'success': False, 'message': error.args[0]}
    else:
        return {'success': True, 'data': config_data}


def export(lims_api, customer_id, family_id):
    lims_samples = lims_api.case(customer_id, family_id)
    case_data = export_case(lims_api, lims_samples)
    return case_data


def invoice(lims_samples):
    """Fetch information about invoicing."""
    customers = set()
    samples = []
    for lims_sample in lims_samples:
        sample_date = parse_date(lims_sample.date_received)
        try:
            customers.add(lims_sample.udf['customer'])
            data = dict(
                lims_id=lims_sample.id,
                date=sample_date.date(),
                application_tag=lims_sample.udf['Sequencing Analysis'],
                application_tag_version=int(lims_sample.udf['Application Tag Version']),
                priority=lims_sample.udf['priority'],
                project=lims_sample.project.name,
                name=lims_sample.name,
            )
        except KeyError as error:
            message = "Missing UDF value: {} - {}".format(lims_sample.id, error.message)
            raise MissingLimsDataException(message)
        samples.append(data)

    if len(customers) > 1:
        raise MultipleCustomersError(customers)
    return dict(samples=samples, customer_id=customers.pop())


def invoice_process(lims_api, process_id):
    """Get information from an invoice process."""
    lims_process = Process(lims_api, id=process_id)
    lims_data = process_samples(lims_process)

    lims_samples = {}
    for sample_data in lims_data:
        if sample_data['artifact'].name.startswith('Pool'):
            lims_artifact = sample_data['artifact']
            if lims_artifact.id not in lims_samples:
                lims_sample = sample_data['sample']
                lims_artifact.date_received = lims_sample.date_received
                lims_artifact.project = lims_sample.project
                udf_keys = ['customer', 'Sequencing Analysis', 'Application Tag Version',
                            'priority']
                for udf_key in udf_keys:
                    lims_artifact.udf[udf_key] = lims_sample.udf[udf_key]
                lims_samples[lims_artifact.id] = lims_artifact
        else:
            lims_samples[sample_data['sample'].id] = sample_data['sample']

    data = dict(
        lims_samples=lims_samples.values(),
        discount=float(lims_process.udf['Discount (%)']),
        lims_process=lims_process,
    )
    return data


def validate(lims_api, customer_id, family_name):
    """Get information used to validate a sample."""
    data = {}
    for lims_sample in lims_api.case(customer_id, family_name):
        sample = api.ClinicalSample(lims_sample)
        data[sample.sample_id] = dict(
            is_external=sample.apptag.is_external,
            is_panel=sample.apptag.is_panel,
            reads=sample.apptag.reads,
            is_low_input=sample.apptag.startswith('WGSLIN'),
        )
    return data


def start(lims_api, customer_id, family_name):
    """Fetch data relevant to starting an analysis."""
    statuses = set()
    is_externals = set()
    for lims_sample in lims_api.case(customer_id, family_name):
        sample = api.ClinicalSample(lims_sample)

        is_external = sample.apptag.is_external
        log.debug("%s is %s external", sample.sample_id, '' if is_external else 'not')
        is_externals.add(is_external)

        sample_status = sample.udf('priority')
        log.debug("%s has status %s", sample.sample_id, sample_status)
        statuses.add(sample_status)

    if len(is_externals) == 1:
        data = dict(is_external=is_externals.pop())
    else:
        # mix of internal and external samples
        data = dict(is_external=False)

    data['is_prio'] = True if ('express' in statuses) or ('priority' in statuses) else False
    return data
